import logging
from dataclasses import dataclass
from dataclasses import field
from multiprocessing import Manager
from multiprocessing import Process
from multiprocessing import Queue
from typing import Any
from typing import Dict
from typing import Tuple


@dataclass
class Task:
    name: str
    args: Tuple[Any, ...] = field(default_factory=tuple)
    kwargs: Dict[str, Any] = field(default_factory=dict)


class BaseQueue:
    def __init__(
        self,
        **kwargs,
    ):
        self.kwargs = kwargs
        self._task_queue = Queue()
        self._worker_process = Process(target=self._worker_function)
        self._worker_process.start()

    def _get_context(self) -> Dict[str, Any]:
        low_priority_queue = Queue()

        return {
            "low_priority_queue": low_priority_queue,
        }

    def enqueue_high_prio(self, task_name, *args, wait_for_result=True, **kwargs):
        result_queue = Manager().Queue()
        task = Task(
            name=task_name, args=args, kwargs={**kwargs, "__result_queue": result_queue}
        )
        self._task_queue.put(task)
        if wait_for_result:
            return result_queue.get()

        return None

    def shutdown(self):
        self._task_queue.put(Task(name="shutdown", args=(), kwargs={}))
        self._worker_process.join()

    def _handle_task(self, context, task: Task):
        logging.info("Handling task: %s", task.name)
        handler_name = f"handle_{task.name}"
        handler = getattr(self, handler_name, None)
        if handler:
            kwargs = dict(task.kwargs or {})
            result_queue = kwargs.pop("__result_queue", None)
            result = handler(context, *task.args, **kwargs)
            if result_queue is not None:
                result_queue.put(result)

    def _worker_function(self):
        logging.info("Starting worker process...")
        context = self._get_context()
        low_priority_queue = context["low_priority_queue"]

        while True:
            while self._task_queue.qsize() == 0 and low_priority_queue.qsize() > 0:
                task = low_priority_queue.get()
                self._handle_task(context, task)

            task = self._task_queue.get()

            if task.name == "shutdown":
                break

            self._handle_task(context, task)



# pylint: disable=too-many-arguments
import os
import sys

import click
import orjson
import requests

from seagoat import __version__
from seagoat.utils.cli_display import display_results
from seagoat.utils.server import get_server_info


# pylint: disable-next=too-few-public-methods
class ExitCode:
    SERVER_NOT_RUNNING = 3
    SERVER_ERROR = 4


def display_accuracy_warning(server_address):
    response = requests.get(
        f"{server_address}/status",
    )
    response_data = orjson.loads(response.text)
    accuracy = response_data["stats"]["accuracy"]["percentage"]

    if accuracy < 100:
        click.echo(
            click.style(
                "Warning: SeaGOAT is still analyzing your repository. "
                + f"The results displayed have an estimated accuracy of {accuracy}%",
                fg="red",
            ),
            err=True,
        )


def query_server(
    query, server_address, repo_path, max_results, context_above, context_below
):
    try:
        response = requests.get(
            f"{server_address}/query/{query}",
            params={
                "limitClue": max_results,
                "contextAbove": context_above,
                "contextBelow": context_below,
            },
        )

        response_data = orjson.loads(response.text)
        if "error" in response_data:
            click.echo(response_data["error"]["message"], err=True)
            sys.exit(ExitCode.SERVER_ERROR)

        response.raise_for_status()
        return response_data["results"]
    except (
        requests.exceptions.ConnectionError,
        requests.exceptions.RequestException,
    ):
        click.echo(
            f"The SeaGOAT server is not running. "
            f"Please start the server using the following command:\n\n"
            f"seagoat-server start {repo_path}\n"
        )
        sys.exit(ExitCode.SERVER_NOT_RUNNING)


@click.command()
@click.argument("query")
@click.argument("repo_path", required=False, default=os.getcwd())
@click.option(
    "--no-color",
    is_flag=True,
    help="Disable formatting. Automatically enabled when part of a bash pipeline.",
)
@click.option(
    "-l",
    "--max-results",
    type=int,
    default=None,
    help="Limit the number of result lines",
)
@click.option(
    "-B",
    "--context-above",
    type=int,
    default=None,
    help="Include this many lines of context before each result",
)
@click.option(
    "-A",
    "--context-below",
    type=int,
    default=None,
    help="Include this many lines of context after each result",
)
@click.option(
    "-C",
    "--context",
    type=int,
    default=None,
    help="Include this many lines of context after and before each result",
)
@click.version_option(version=__version__, prog_name="seagoat")
def seagoat(
    query, repo_path, no_color, max_results, context_above, context_below, context
):
    """
    Query your codebase for your QUERY in the Git repository REPO_PATH.
    Your query can contain keywords, regular expression patterns,
    or a description of what you are looking for.

    When REPO_PATH is not specified, the current working directory is
    assumed to be the repository path.

    In order to use seagoat in your repository, you need to run a server
    that will analyze your codebase. Check seagoat-server --help for more details.
    """
    server_info = get_server_info(repo_path)
    server_address = server_info["address"]

    if context is not None:
        context_above = context
        context_below = context

    results = query_server(
        query,
        server_address,
        repo_path,
        max_results,
        context_above or 0,
        context_below or 0,
    )

    color_enabled = os.isatty(0) and not no_color

    display_results(results, max_results, color_enabled)

    display_accuracy_warning(server_address)


if __name__ == "__main__":
    # pylint: disable-next=no-value-for-parameter
    seagoat()



import os
import socket

import appdirs

from seagoat.utils.json_file import get_json_file_contents


def get_server_info_file(repo_path):
    repo_id = os.path.normpath(repo_path).replace(os.sep, "_")
    user_cache_dir = appdirs.user_cache_dir("seagoat-servers")
    os.makedirs(user_cache_dir, exist_ok=True)
    return os.path.join(user_cache_dir, f"{repo_id}.json")


def get_server_info(repo_path: str):
    server_info = get_json_file_contents(get_server_info_file(repo_path))
    server_info["address"] = f"http://{server_info['host']}:{server_info['port']}"

    return server_info


def is_server_running(repo_path: str):
    server_info = get_server_info(repo_path)
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as socket_obj:
        return socket_obj.connect_ex((server_info["host"], server_info["port"])) == 0


def get_free_port():
    socket_obj = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    socket_obj.bind(("", 0))
    _, port = socket_obj.getsockname()
    socket_obj.close()

    return port



import time


def wait_for(condition_function, timeout, period=0.05):
    start_time = time.time()
    while not condition_function():
        if time.time() - start_time > timeout:
            raise TimeoutError("Timeout expired while waiting for condition.")
        time.sleep(period)



from pathlib import Path

import chromadb
from chromadb.config import Settings
from chromadb.errors import IDAlreadyExistsError

from seagoat.cache import Cache
from seagoat.repository import Repository
from seagoat.result import Result


def initialize(repository: Repository):
    cache = Cache("chroma", Path(repository.path), {})

    chroma_client = chromadb.PersistentClient(
        path=str(cache.get_cache_folder()),
        settings=Settings(
            anonymized_telemetry=False,
        ),
    )

    chroma_collection = chroma_client.get_or_create_collection(name="code_data")

    def fetch(query_text: str, limit: int):
        # Slightly overfetch results as it will sorted using a different score later
        maximum_chunks_to_fetch = 100  # this should be plenty, especially because many times context could be included
        worst_distance_to_display = 1.5
        n_results = min((limit + 1) * 2, maximum_chunks_to_fetch)

        chromadb_results = [
            chroma_collection.query(
                query_texts=[query_text],
                n_results=n_results,
            )
        ]
        metadata_with_distance = (
            list(
                zip(
                    chromadb_results[0]["metadatas"][0],
                    chromadb_results[0]["distances"][0],
                )
            )
            if chromadb_results[0]["metadatas"] and chromadb_results[0]["distances"]
            else None
        ) or []

        files = {}

        for metadata, distance in metadata_with_distance:
            if distance > worst_distance_to_display:
                break
            path = str(metadata["path"])
            line = int(metadata["line"])

            if path not in files:
                files[path] = Result(path, Path(repository.path) / path)
            files[path].add_line(line, distance)

        return files.values()

    def cache_chunk(chunk):
        try:
            chroma_collection.add(
                ids=[chunk.chunk_id],
                documents=[chunk.chunk],
                metadatas=[{"path": chunk.path, "line": chunk.codeline}],
            )
        except IDAlreadyExistsError:
            pass

    return {
        "fetch": fetch,
        "cache_chunk": cache_chunk,
    }



# pylint: disable=import-outside-toplevel
import math
import subprocess
from functools import cache
from typing import Optional

import click


@cache
def get_highlighted_lines(file_name: str):
    from pygments import highlight
    from pygments.formatters import TerminalFormatter
    from pygments.lexers import get_lexer_for_filename
    from pygments.lexers.javascript import JavascriptLexer
    from pygments.lexers.javascript import TypeScriptLexer

    with open(file_name, "r", encoding="utf-8") as source_code_file:
        code = source_code_file.read()

    if file_name.endswith(".md"):
        return code.splitlines()

    if file_name.endswith(".jsx"):
        lexer = JavascriptLexer()
    elif file_name.endswith(".tsx"):
        lexer = TypeScriptLexer()
    else:
        lexer = get_lexer_for_filename(file_name)

    result = highlight(code, lexer, TerminalFormatter())

    return result.splitlines()


def print_result_line(result, block, line, color_enabled):
    if color_enabled:
        highlighted_lines = get_highlighted_lines(str(result["fullPath"]))
        click.echo(
            f"{result['path']}:{click.style(str(line), bold=True)}:{highlighted_lines[line - 1]}",
            color=True,
        )
    else:
        for line_content in block["lines"]:
            if line_content["line"] == line:
                click.echo(f"{result['path']}:{line}:{line_content['lineText']}")
                break


def iterate_result_blocks(results, max_results: Optional[int]):
    lines_left_to_print = max_results if max_results is not None else math.inf
    number_of_blocks_printed = 0

    for result in results:
        for block in result["blocks"]:
            if lines_left_to_print <= 0 and number_of_blocks_printed >= 1:
                return

            lines_left_to_print -= block["lineTypeCount"]["result"]

            yield result, block

            number_of_blocks_printed += 1


def display_results_using_bat(results, max_results):
    current_result = None
    blocks = []

    for result, block in iterate_result_blocks(results, max_results):
        if current_result is None:
            current_result = result

        if result["fullPath"] != current_result["fullPath"]:
            display_blocks_with_bat(current_result, blocks)
            current_result = result
            blocks = []

        blocks.append(block)

    display_blocks_with_bat(current_result, blocks)


def display_results(results, max_results, color_enabled):
    if color_enabled and is_bat_installed():
        display_results_using_bat(results, max_results)
        return

    for result, block in iterate_result_blocks(results, max_results):
        print_result_block(result, block, color_enabled)


def is_bat_installed():
    try:
        result = subprocess.run(
            ["bat", "--version"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )
        return result.returncode == 0
    except FileNotFoundError:
        return False


def display_blocks_with_bat(result, blocks):
    file_name = result["path"]
    full_path = result["fullPath"]

    command = [
        "bat",
        full_path,
        "--file-name",
        file_name,
        "--paging",
        "never",
    ]

    for block in blocks:
        start_line_number = block["lines"][0]["line"]
        end_line_number = block["lines"][-1]["line"]
        command.extend(["--line-range", f"{start_line_number}:{end_line_number}"])

    subprocess.run(command, check=True)


def print_result_block(result, block, color_enabled):
    for result_line in block["lines"]:
        line = result_line["line"]
        print_result_line(result, block, line, color_enabled)
